/*
 * Copyright (C) 2007 - John Anderson, Neil Kettle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#ifdef DEBUG
#include <assert.h>
#endif
#include <getopt.h>
#include <time.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#include "common.h"
#include "net.h"
#include "obuf.h"

#include "synscan.h"

/* Global variables */
char net_interface[MAX_IFNAMELEN];      /* network interface */
char *arg_progname;                     /* program name      */

static u_int32_t src_ip;

/* Statistics variables */
static time_t start_t;
static unsigned long long num_port = 0;
static unsigned long long num_ip = 0;
static unsigned long long total_scan = 0;

/* Static variables */
static const struct option syn_options[] = {
  /*
   * Options
   */
  {"block", required_argument, 0, 'b'}, /* IP block */
  {"file", required_argument, 0, 'f'},  /* IP file */
  {"port", required_argument, 0, 'p'},  /* port block */
  {"interface", required_argument, 0, 'I'},     /* interface */
  {"burst", required_argument, 0, 'B'}, /* packet burst length */
  {"time", required_argument, 0, 'T'},  /* time between bursts */
  {"rand-time", no_argument, 0, 256},   /* time variance */

  {"flags", required_argument, 0, 512}, /* IP flags */
  {"rand-flags", no_argument, 0, 1024}, /* randomise IP flags */
  {"src-port", required_argument, 0, 2048},     /* src port */
  {"rand-src-port", no_argument, 0, 4096},      /* randomise src port */

  /*
   * Misc.
   */
  {"version", no_argument, 0, 'V'},     /* version */
  {"verbose", no_argument, 0, 'v'},     /* verbose output */
  {"help", no_argument, 0, 'h'},        /* print usage */
  {NULL, 0, 0, 0}
};

static int arg_iplist;
static char arg_ipliststr[MAX_ARGSIZE];
static char arg_portliststr[MAX_ARGSIZE];

static int arg_burst;
static int arg_delay;
static int arg_verbose;

static void
print_header (void)
{
  fprintf (stdout, "SynScan " SYNSCAN_VERSION " (" SYNSCAN_URL ")\n"
           "by John Anderson <john [at] ev6.net>,\n"
           "   Neil Kettle <mu-b [at] digit-labs.org>.\n\n");
}

static void
print_usage (char **argv)
{
  struct option *cur = (struct option *) syn_options;

  fprintf (stdout, "Usage: %s [options]\n\n", argv[0]);

  do
    {
      switch (cur->val)
        {
        case 'b':
          fprintf (stdout, "-%c --%-15s\tspecify network block to scan\n",
                   cur->val, cur->name);
          break;
        case 'f':
          fprintf (stdout,
                   "-%c --%-15s\tspecify network block file to scan\n",
                   cur->val, cur->name);
          break;
        case 'p':
          fprintf (stdout, "-%c --%-15s\tspecify port block scan [%s]\n",
                   cur->val, cur->name, SYNSCAN_DEFPORTS);
          break;
        case 'B':
          fprintf (stdout, "-%c --%-15s\tspecify packet burst length [%d]\n",
                   cur->val, cur->name, SYNSCAN_DEFBURST);
          break;
        case 'T':
          fprintf (stdout,
                   "-%c --%-15s\tspecify packet send delay between bursts (usec, nX msec)\n"
                   "\t\t\tdefault: %d packets are sent, with delay %d usec\n",
                   cur->val, cur->name, SYNSCAN_DEFBURST, SYNSCAN_DEFDELAY);
          break;
        case 256:
          fprintf (stdout, "--%-15s\trandomise packet send delay\n",
                   cur->name);
          break;
        case 'I':
          fprintf (stdout, "-%c --%-15s\tspecify network interface [%s]\n",
                   cur->val, cur->name, SYNSCAN_DEFIFC);
          break;
        case 512:
          fprintf (stdout, "--%-15s\tspecify IP flags\n", cur->name);
          break;
        case 1024:
          fprintf (stdout, "--%-15s\trandomise IP flags\n", cur->name);
          break;
        case 2048:
          fprintf (stdout, "--%-15s\tspecify TCP source port (=dest port)\n",
                   cur->name);
          break;
        case 4096:
          fprintf (stdout, "--%-15s\trandomise TCP source port\n", cur->name);
          break;

          /* the rest... */
        case 'V':
          fprintf (stdout, "-%c --%-15s\tprint version information\n",
                   cur->val, cur->name);
          break;
        case 'v':
          fprintf (stdout, "-%c --%-15s\tverbose output\n",
                   cur->val, cur->name);
          break;
        case 'h':
          fprintf (stdout, "-%c --%-15s\tprint this message\n",
                   cur->val, cur->name);
          break;
        }
    }
  while ((++cur)->name);
}

static void
syn_hostportscan (struct net_pkt_t *pkt, u_int32_t cur_ip, u_int16_t cur_port)
{
  static struct stats_t
  {
    time_t last_t;
    unsigned long long done;
  } stats = { 0, 0 };
  static int cur_burst = 0;
  time_t tmp_t;

  cur_burst++;
  if (cur_burst == arg_burst)
    {
      cur_burst = 0;
      usleep (arg_delay);
    }

  net_ipv4_send (pkt, src_ip, cur_ip, cur_port, cur_port);
  stats.done++;

  if (time (&tmp_t) > stats.last_t)
    {
      double cur_percent;
      char *cur_ip_str;

      stats.last_t = tmp_t;
      cur_ip_str = nlookup (cur_ip);
      cur_percent = (((double) stats.done) / total_scan) * 100;
      fprintf (stdout,
               "| %15s | %5d | %4d | %6llu | %3.0f | %15llu | %15llu |\r",
               cur_ip_str, cur_port, (int) (stats.last_t - start_t),
               stats.done / (stats.last_t >
                             start_t ? stats.last_t - start_t : stats.done),
               cur_percent, stats.done, total_scan - stats.done);
      fflush (stdout);
    }
}

static void
parse_args (int argc, char **argv)
{
  extern char *optarg;
  char *opt_tmp;
  int ch, opt_index, opt_len;

  if (argc <= 1)
    {
      print_usage (argv);
      exit (EXIT_SUCCESS);
    }

  /* initialise argument globals */
  arg_progname = argv[0];

  arg_iplist = -1;
  memset (arg_ipliststr, 0, sizeof arg_ipliststr);

  memset (arg_portliststr, 0, sizeof arg_portliststr);
  snprintf (arg_portliststr, sizeof arg_portliststr, SYNSCAN_DEFPORTS);
  memset (net_interface, 0, sizeof net_interface);
  snprintf (net_interface, sizeof net_interface, SYNSCAN_DEFIFC);

  arg_burst = SYNSCAN_DEFBURST;
  arg_delay = SYNSCAN_DEFDELAY;
  arg_verbose = 0;

  while ((ch = getopt_long (argc, argv, "b:f:p:I:B:T:Vvh", syn_options,
                            &opt_index)) != -1)
    {
      switch (ch)
        {
        case 'b':
        case 'f':
          opt_len = strlen (optarg);
          if (opt_len > sizeof arg_ipliststr - 1)
            {
              warning ("%s: IP block%s to long (>%d-bytes)\n",
                       arg_progname, (ch == 'b' ? ""
                                      : " file name"), sizeof arg_ipliststr);
              goto failed;
            }

          arg_iplist = (ch == 'b' ? IPLIST_BLOCK : IPLIST_FILE);
          strncpy (arg_ipliststr, optarg, sizeof arg_ipliststr - 1);
          arg_ipliststr[sizeof arg_ipliststr - 1] = '\0';
          break;

        case 'p':
          opt_len = strlen (optarg);
          if (opt_len > sizeof arg_portliststr - 1)
            {
              warning ("%s: port block to long (>%d-bytes)\n",
                       arg_progname, sizeof arg_portliststr);
              goto failed;
            }

          strncpy (arg_portliststr, optarg, sizeof arg_portliststr - 1);
          arg_portliststr[sizeof arg_portliststr - 1] = '\0';
          break;

        case 'I':
          strncpy (net_interface, optarg, sizeof net_interface - 1);
          net_interface[sizeof net_interface - 1] = '\0';
          break;

        case 'B':
          arg_burst = strtol (optarg, &opt_tmp, 10);
          if (optarg[0] == '+' || optarg[0] == '-' || *opt_tmp != '\0')
            {
              warning ("%s: invalid burst length\n", arg_progname);
              goto failed;
            }
          break;

        case 'T':
          {
            int is_msec;

            is_msec = 0;
            opt_len = strlen (optarg);

            if (optarg[opt_len - 1] == 'X')
              is_msec = 1, optarg[opt_len - 1] = '\0';

            arg_delay = strtol (optarg, &opt_tmp, 10);
            if (optarg[0] == '+' || optarg[0] == '-' || *opt_tmp != '\0')
              {
                warning ("%s: invalid delay time\n", arg_progname);
                goto failed;
              }

            if (is_msec)
              arg_delay *= 1000;
          }
          break;

        case 'V':
          exit (EXIT_SUCCESS);
          break;
        case 'v':
          arg_verbose++;
          break;
        case 'h':
          print_usage (argv);
        default:
          exit (EXIT_FAILURE);
        }
    }

  if (arg_iplist == -1)
    {
      warning ("%s: no hosts to scan!\n", arg_progname);
      goto failed;
    }

  return;

failed:
  /* at least one argument is incorrect.. */
  fatal ("Try `%s -h' for more information\n", arg_progname);
}

int
main (int argc, char **argv)
{
  struct iplistv4_t *iplist;
  struct portlist_t *portlist;
  struct iplistv4_iter_t iplist_iter;
  struct portlist_iter_t portlist_iter;
  struct net_pkt_t net_pkt;
  time_t total_t;

  /* print the header and parse the arguments */
  print_header ();
  parse_args (argc, argv);

  /* check for root privs */
  if (getuid () && geteuid ())
    fatal ("%s: getuid(): UID or EUID of 0 required\n", arg_progname);

  /* parse the ip lists */
  iplist = NULL;
  if (arg_iplist == IPLIST_BLOCK)
    iplist = net_parse_iplistv4 (arg_ipliststr);
  else
    {
      if (arg_iplist == IPLIST_FILE)
        {
          char *filen;
          FILE *filen_fp;

          filen = arg_ipliststr;
          if ((filen_fp = fopen (filen, "r")) == NULL)
            fatal ("%s: cannot open ip list file '%s'\n", arg_progname,
                   filen);

          iplist = net_parse_iplistv4_file (filen, filen_fp);
        }
      else
        {
          fatal ("%s: ip list is not a block nor file! (impossible)\n",
                 arg_progname);
        }
    }

  if (iplist == NULL)
    fatal ("%s: failed parsing ip %s\n", arg_progname,
           (arg_iplist == IPLIST_BLOCK ? "block" : "file"));

  /* parse the port lists */
  portlist = net_parse_portlist (arg_portliststr);
  if (portlist == NULL)
    fatal ("%s: failed parsing port list\n", arg_progname);

  /* get our interfaces IP */
  if (net_get_localip (net_interface, &src_ip) == -1)
    fatal ("%s: no ip address found for device %s (non-assigned, or interface down)\n",
           arg_progname, net_interface);

  /* calculate the total number of IPs/ports */
  num_ip = net_iplistv4_sum (iplist);
  num_port = net_portlist_sum (portlist);
  total_scan = num_ip * num_port;

  start_t = time (NULL);

  {
    struct in_addr cur_ip;
    sock_t rsock;

    /* initialise a raw socket */
    rsock = net_sock_init ();

    /* initialise our packet */
    net_ipv4_pkt_init (&net_pkt, rsock, TH_SYN);

    net_iplistv4_iter_init (&iplist_iter, iplist);
    cur_ip = iplist_iter.net_addr_t.ip;

    fprintf (stdout, "Scanning networks:\n");
    net_iplistv4_print (iplist);
    fprintf (stdout, "on ports:\n");
    net_portlist_print (portlist);
    fprintf (stdout, "from IP: %s\n\n", nlookup (src_ip));

    fprintf (stdout,
             "-------------------------------------------------------------------------------------\n"
             "| IP              | PORT  | TIME | SPEED  |  %%  | DONE            | REMAINING       |\n");

    do
      {
        u_int16_t cur_port;

        net_portlist_iter_init (&portlist_iter, portlist);
        cur_port = portlist_iter.cur_port;

        do
          {
#ifdef DEBUG
            printf ("%s:%d\n", nlookup (cur_ip.s_addr), cur_port);
#endif
            syn_hostportscan (&net_pkt, cur_ip.s_addr, cur_port);
          }
        while (net_portlist_iter_next (&portlist_iter, &cur_port) ==
               ITER_CONTINUE);
      }
    while (net_iplistv4_iter_next (&iplist_iter, &cur_ip) == ITER_CONTINUE);
  }

  fprintf (stdout,
           "\n------------------------------------------------------------------------------------\n");

  total_t = time (NULL) - start_t;
  if (total_t > 0)
    fprintf (stdout,
             "%llu IPs and %llu ports scanned in %d seconds, avg %llu p/sec\n",
             num_ip, num_port, (int) total_t, total_scan / total_t);
  else
    fprintf (stdout,
             "%llu IPs and %llu ports scanned in <1 second, avg %llu p/sec\n",
             num_ip, num_port, total_scan);

  net_ipv4_pkt_free (&net_pkt);
  net_iplistv4_free (iplist);
  net_portlist_free (portlist);

  return (EXIT_SUCCESS);
}
